"""File with ridge regression model algorithm."""

# Third-party libraries
import numpy as np

# First-party libraries
from pamm.algorithms.regression.regression_model import RegressionAlgorithm


class ElasticNetRegressionAlgorithm(RegressionAlgorithm):
    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
        alpha: float = 0.01,
        ratio: float = 0.5,
        eta: float = 0.1,
        n_epochs: int = 10,
        t0: int = 1,
        t1: int = 10,
    ) -> np.ndarray:
        x = self._add_bias(x)
        m, n = x.shape
        b = np.random.randn(n, 1)
        for epoch in range(n_epochs):
            for i in range(m):
                random_index = np.random.randint(m)
                xi = x[random_index : random_index + 1]
                yi = y[random_index : random_index + 1]
                gradients = (
                    2 * xi.T @ (xi @ b - yi)
                    + ratio * alpha * sum(abs(b))[0]
                    + (1 - ratio) / 2 * alpha * (b.T @ b)[0, 0]
                )
                eta = t0 / ((epoch * m + i) + t1)
                b = b - eta * gradients

        self._beta = b[int(self._fit_bias) :]
        self._bias = b[0][0] * int(self._fit_bias)



# Standard Library
from typing import Literal

# Third-party libraries
import numpy as np


def forward(major_ts: np.ndarray, minor_ts: np.ndarray, trust: int) -> np.ndarray:
    """Reterned values from timeseries nearest forward in time.

    Args:
        major_ts (np.ndarray): time-series original.
        minor_ts (np.ndarray): time-series for merge.
        trust (int): trust interval.

    Returns:
        np.ndarray: merged values.
    """
    arr = np.empty(major_ts.shape)
    indexes = np.searchsorted(minor_ts[:, 0], major_ts[:, 0], side="left")
    _tmp = np.atleast_2d(np.take(minor_ts[:, 0], [indexes], mode="wrap") - major_ts[:, 0]).T
    arr[:, 0] = np.take(minor_ts[:, 0], [indexes], mode="wrap")
    arr[:, 1] = np.take(minor_ts[:, 1], [indexes], mode="wrap")
    arr = np.hstack((_tmp,arr))
    arr = arr.astype(float)
    arr[arr[:, 0] > trust] = np.nan
    arr[arr[:, 0] < 0] = np.nan
    return arr[:,1:]


def backward(major_ts: np.ndarray, minor_ts: np.ndarray, trust: int) -> np.ndarray:
    """Reterned values from timeseries nearest back in time.

    Args:
        major_ts (np.ndarray): time-series original.
        minor_ts (np.ndarray): time-series for merge.
        trust (int): trust interval.

    Returns:
        np.ndarray: merged values.
    """
    arr = np.empty(major_ts.shape)
    indexes = np.searchsorted(minor_ts[:, 0], major_ts[:, 0], side="right") - 1
    _tmp = np.atleast_2d(major_ts[:, 0] - np.take(minor_ts[:, 0], [indexes], mode="wrap")).T
    arr[:, 0] = np.take(minor_ts[:, 0], [indexes], mode="wrap")
    arr[:, 1] = np.take(minor_ts[:, 1], [indexes], mode="wrap")
    arr = np.hstack((_tmp,arr))
    arr = arr.astype(float)
    arr[arr[:, 0] > trust] = np.nan
    arr[arr[:, 0] < 0] = np.nan
    return arr[:,1:]


def nearest(major_ts: np.ndarray, minor_ts: np.ndarray, trust: int) -> np.ndarray:
    """Reterned values from timeseries nearest in time.

    Args:
        major_ts (np.ndarray): time-series original.
        minor_ts (np.ndarray): time-series for merge.
        trust (int): trust interval.

    Returns:
        np.ndarray: merged values.
    """
    indexes = np.searchsorted(minor_ts[:, 0], major_ts[:, 0], "right")
    left_diff = np.abs(
        np.take(minor_ts[:, 0], indexes - 1, mode="wrap") - major_ts[:, 0]
    )
    right_diff = np.abs(np.take(minor_ts[:, 0], indexes, mode="wrap") - major_ts[:, 0])
    nearest_indexes = np.where(left_diff > right_diff, indexes, indexes - 1)
    arr = minor_ts[nearest_indexes]
    _tmp = np.atleast_2d(np.abs(arr[:, 0] - major_ts[:, 0])).T
    arr = np.hstack((_tmp,arr))
    arr = arr.astype(float)
    arr[arr[:, 0] > trust] = np.nan
    return arr[:,1:]


def sync_ts(
    major_ts: np.ndarray,
    minor_ts: np.ndarray,
    method: Literal["backward", "forward", "nearest"] = "backward",
    trust: int = np.inf,
) -> np.ndarray:
    """_summary_

    Args:
        major_ts (np.ndarray): _description_
        minor_ts (np.ndarray): _description_
        method (Literal[backward, forward, nearest], optional): _description_. Defaults to 'backward'.
        trust (int, optional): _description_. Defaults to np.inf.

    Returns:
        np.ndarray: merged data
    """

    match method:
        case "backward":
            return backward(major_ts=major_ts, minor_ts=minor_ts, trust=trust)

        case "forward":
            return forward(major_ts=major_ts, minor_ts=minor_ts, trust=trust)

        case "nearest":
            return nearest(major_ts=major_ts, minor_ts=minor_ts, trust=trust)


"""File with laso regression model algorithm."""

# Third-party libraries
import numpy as np

# First-party libraries
from pamm.algorithms.regression.regression_model import RegressionAlgorithm


class LassoRegressionAlgorithm(RegressionAlgorithm):
    def _sign(self, num: float) -> float:
        if num == 0:
            return num
        if num > 0:
            return 1
        if num < 0:
            return -1

    def _sign_vectorize(self, b: np.ndarray) -> np.ndarray:
        vector = np.zeros(b.shape)
        for x in range(b.shape[0]):
            vector[x, 0] = self._sign(b[x, 0])
        return vector

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
        alpha: float = 0.01,
        eta: float = 0.1,
        n_epochs: int = 10,
        t0: int = 1,
        t1: int = 10,
    ) -> np.ndarray:
        x = self._add_bias(x)
        m, n = x.shape
        b = np.random.randn(n, 1)
        for epoch in range(n_epochs):
            for i in range(m):
                random_index = np.random.randint(m)
                xi = x[random_index : random_index + 1]
                yi = y[random_index : random_index + 1]
                gradients = 2 * xi.T @ (xi @ b - yi) + alpha * self._sign_vectorize(b)
                eta = t0 / ((epoch * m + i) + t1)
                b = b - eta * gradients

        self._beta = b[int(self._fit_bias) :]
        self._bias = b[0][0] * int(self._fit_bias)



# Third-party libraries
import numpy as np

# First-party libraries
from pamm.algorithms.base_model import AbstractBaseModelAlgorithm


def cv(
    model: AbstractBaseModelAlgorithm, x: np.array, y: np.array, cv: int = 2
) -> None:
    len_ds = int(np.ceil(x.shape[0] / cv))
    models = [model for _ in range(cv)]

    for num, model in enumerate(models):
        x_reduce = x[num * len_ds : (num + 1) * len_ds]
        y_reduce = y[num * len_ds : (num + 1) * len_ds]
        model.fit(x=x_reduce, y=y_reduce)



# First-party libraries
from pamm.algorithms.regression.gradient_boost.pgd_regression import \
    BatchGradientRegressionAlgorithm
from pamm.algorithms.regression.gradient_boost.sgd_regression import \
    StachosticGradientRegressionAlgorithm
from pamm.algorithms.regression.iterations_regression import \
    IterationRegressionAlgorithm
from pamm.algorithms.regression.jackknife_model import \
    JackKnifeRegressionAlgorithm
from pamm.algorithms.regression.qp_mls_model import QPMLSAlgorithm
from pamm.algorithms.regression.regression_model import RegressionAlgorithm
from pamm.algorithms.regression.regularization.elastic_net import \
    ElasticNetRegressionAlgorithm
from pamm.algorithms.regression.regularization.lasso_model import \
    LassoRegressionAlgorithm
from pamm.algorithms.regression.regularization.ridge_model import \
    RidgeRegressionAlgorithm
from pamm.algorithms.regression.tsls_model import TwoStageLeastSquaresAlgorithm
from pamm.algorithms.regression.wls.mls_model import ModifyRegressionAlgorithm
from pamm.algorithms.regression.wls.qp_model import QPRegressionAlgorithm
from pamm.algorithms.regression.wls.wls_model import \
    WeigthedRegressionAlgorithm
from pamm.algorithms.regression.wls.wtsls_model import \
    WeightedTwoStageLeastSquaresAlgorithm



"""File with regression model algorithm on two-stage least sqares."""

# Third-party libraries
import numpy as np

# First-party libraries
from pamm.algorithms.regression.regression_model import RegressionAlgorithm


class JackKnifeRegressionAlgorithm(RegressionAlgorithm):
    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of endogenous regressors (features)
            y (np.ndarray): matrix a teacher
        """
        x = self._add_bias(x)

        betas = np.zeros(x.T.shape)

        for i in range(x.shape[0]):
            x_reduce = np.delete(x, i, axis=0)
            y_reduce = np.delete(y, i, axis=0)
            betas[:, 0] = (
                np.linalg.inv(x_reduce.T @ x_reduce) @ x_reduce.T @ y_reduce
            ).T

        tmp = np.atleast_2d(np.mean(betas, axis=1)).T

        self._beta = tmp[int(self._fit_bias) :]
        self._bias = tmp[0][0] * int(self._fit_bias)



"""File with regression model algorithm on two-stage least sqares."""

# Third-party libraries
import numpy as np

# First-party libraries
from pamm.algorithms.regression.regression_model import RegressionAlgorithm


class BatchGradientRegressionAlgorithm(RegressionAlgorithm):
    def fit(
        self, x: np.ndarray, y: np.ndarray, eta: float = 0.1, max_iter: int = 1000
    ) -> np.ndarray:
        x = self._add_bias(x)
        m, n = x.shape
        theta = np.random.randn(n, 1)
        for iteration in range(max_iter):
            gradients = 2 / m * x.T @ (x @ theta - y)
            theta = theta - eta * gradients

        self._beta = theta[int(self._fit_bias) :]
        self._bias = theta[0][0] * int(self._fit_bias)
